package io.zipcoder.crudapp;

import com.fasterxml.jackson.databind.ObjectMapper;
import io.zipcoder.crudapp.controllers.PersonController;
import io.zipcoder.crudapp.models.Person;
import io.zipcoder.crudapp.services.PersonService;
import org.apache.catalina.filters.CorsFilter;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.context.web.WebAppConfiguration;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import java.util.ArrayList;
import java.util.List;

import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.mockito.Mockito.*;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.delete;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;


// ========== Looked further into the readme. Testing Controller the 'right' way now =========//

@WebAppConfiguration
@SpringBootTest
@RunWith(SpringJUnit4ClassRunner.class)
public class PersonControllerTestMVC {
    List<Person> list;
    Person testPerson = new Person("Test", "Person");
    private MockMvc mockMvc;

    @Mock
    private PersonService personService;

    @InjectMocks
    private PersonController personController;

    @Before
    public void init(){
        list = new ArrayList<>();
        list.add(testPerson);

        MockitoAnnotations.initMocks(this);
        mockMvc = MockMvcBuilders
                .standaloneSetup(personController)
                .build();
    }

    // =================== Get All Users =================== //

    @Test
    public void testGetPersonList() throws Exception {
        when(personService.getPersonList()).thenReturn(list);

        mockMvc.perform(get("/people"))
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON_UTF8_VALUE))
                .andExpect(jsonPath("$", hasSize(1)))
                .andExpect(jsonPath("$[0].firstName", is("Test")))
                .andExpect(jsonPath("$[0].lastName",is("Person")));

        verify(personService,times(1)).getPersonList();
        verifyNoMoreInteractions(personService);
    }

    // ======================== Get User By ID ======================== //

    @Test
    public void testGetPersonById() throws Exception {
        when(personService.getPersonById(1L)).thenReturn(testPerson);

        mockMvc.perform(get("/people/{id}",1))
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON_UTF8_VALUE))
                .andExpect(jsonPath("$.firstName",is("Test")))
                .andExpect(jsonPath("$.lastName",is("Person")));

        verify(personService,times(1)).getPersonById(1L);
        verifyNoMoreInteractions(personService);
    }

    @Test
    public void testGetPersonById_fail_404_notFound() throws Exception {
        when(personService.getPersonById(2L)).thenReturn(null);

        mockMvc.perform(get("/people/{id}",2))
                .andExpect(status().isNotFound());

        verify(personService, times(1)).getPersonById(2L);
        verifyNoMoreInteractions(personService);
    }

    // ======================== Create Person ======================== //

    @Test
    public void testCreatePerson() throws Exception {
        when(personService.savePerson(testPerson)).thenReturn(testPerson);

        mockMvc.perform(post("/people")
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(asJsonString(testPerson)))
                .andExpect(status().isCreated());

        verify(personService,times(1)).savePerson(testPerson);
        verifyNoMoreInteractions(personService);
    }

    // ======================== Update Person ======================== //

    @Test
    public void testUpdatePerson() throws Exception {
        when(personService.savePerson(testPerson)).thenReturn(testPerson);

        mockMvc.perform(put("/people/{id}",1)
                .contentType(MediaType.APPLICATION_JSON)
                .content(asJsonString(testPerson)))
                .andExpect(status().isOk());

        verify(personService, times(1)).savePerson(testPerson);
        verifyNoMoreInteractions(personService);
    }

    // ======================== Delete Person ======================== //

    @Test
    public void testDeletePerson() throws Exception {
        doNothing().when(personService).deletePerson(1L);

        mockMvc.perform(delete("/people/{id}",1))
                .andExpect(status().isNoContent());

        verify(personService, times(1)).deletePerson(1L);
        verifyNoMoreInteractions(personService);
    }


    public static String asJsonString(final Object obj) {
        try {
            final ObjectMapper mapper = new ObjectMapper();
            return mapper.writeValueAsString(obj);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
}
